# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A demonstration of toil. Sorts the lines of a file into ascending order by doing a parallel merge sort.
This is an intentionally buggy version that doesn't include restart() for testing purposes.
"""
import codecs
import os
import random
import shutil

from configargparse import ArgumentParser

from toil.common import Toil
from toil.job import Job
from toil.realtimeLogger import RealtimeLogger

defaultLines = 1000
defaultLineLen = 50
sortMemory = '600M'


def setup(job, inputFile, N, downCheckpoints, options):
    """
    Sets up the sort.
    Returns the FileID of the sorted file
    """
    RealtimeLogger.info("Starting the merge sort")
    return job.addChildJobFn(down,
                             inputFile, N, 'root',
                             downCheckpoints,
                             options = options,
                             preemptible=True,
                             memory=sortMemory).rv()


def down(job, inputFileStoreID, N, path, downCheckpoints, options, memory=sortMemory):
    """
    Input is a file, a subdivision size N, and a path in the hierarchy of jobs.
    If the range is larger than a threshold N the range is divided recursively and
    a follow on job is then created which merges back the results else
    the file is sorted and placed in the output.
    """

    RealtimeLogger.info("Down job starting: %s" % path)

    # Read the file
    inputFile = job.fileStore.readGlobalFile(inputFileStoreID, cache=False)
    length = os.path.getsize(inputFile)
    if length > N:
        # We will subdivide the file
        RealtimeLogger.critical("Splitting file: %s of size: %s"
                % (inputFileStoreID, length))
        # Split the file into two copies
        midPoint = getMidPoint(inputFile, 0, length)
        t1 = job.fileStore.getLocalTempFile()
        with open(t1, 'w') as fH:
            fH.write(copySubRangeOfFile(inputFile, 0, midPoint+1))
        t2 = job.fileStore.getLocalTempFile()
        with open(t2, 'w') as fH:
            fH.write(copySubRangeOfFile(inputFile, midPoint+1, length))
        # Call down recursively. By giving the rv() of the two jobs as inputs to the follow-on job, up,
        # we communicate the dependency without hindering concurrency.
        result = job.addFollowOnJobFn(up,
                                    job.addChildJobFn(down, job.fileStore.writeGlobalFile(t1), N, path + '/0',
                                                      downCheckpoints, checkpoint=downCheckpoints, options=options,
                                                      preemptible=True, memory=options.sortMemory).rv(),
                                    job.addChildJobFn(down, job.fileStore.writeGlobalFile(t2), N, path + '/1',
                                                      downCheckpoints, checkpoint=downCheckpoints, options=options,
                                                      preemptible=True, memory=options.mergeMemory).rv(),
                                    path + '/up', preemptible=True, options=options, memory=options.sortMemory).rv()
    else:
        # We can sort this bit of the file
        RealtimeLogger.critical("Sorting file: %s of size: %s"
                % (inputFileStoreID, length))
        # Sort the copy and write back to the fileStore
        shutil.copyfile(inputFile, inputFile + '.sort')
        sort(inputFile + '.sort')
        result = job.fileStore.writeGlobalFile(inputFile + '.sort')

    RealtimeLogger.info("Down job finished: %s" % path)
    return result


def up(job, inputFileID1, inputFileID2, path, options, memory=sortMemory):
    """
    Merges the two files and places them in the output.
    """

    RealtimeLogger.info("Up job starting: %s" % path)

    with job.fileStore.writeGlobalFileStream() as (fileHandle, outputFileStoreID):
        fileHandle = codecs.getwriter('utf-8')(fileHandle)
        with job.fileStore.readGlobalFileStream(inputFileID1) as inputFileHandle1:
            inputFileHandle1 = codecs.getreader('utf-8')(inputFileHandle1)
            with job.fileStore.readGlobalFileStream(inputFileID2) as inputFileHandle2:
                inputFileHandle2 = codecs.getreader('utf-8')(inputFileHandle2)
                RealtimeLogger.info("Merging %s and %s to %s"
                    % (inputFileID1, inputFileID2, outputFileStoreID))
                merge(inputFileHandle1, inputFileHandle2, fileHandle)
        # Cleanup up the input files - these deletes will occur after the completion is successful.
        job.fileStore.deleteGlobalFile(inputFileID1)
        job.fileStore.deleteGlobalFile(inputFileID2)

        RealtimeLogger.info("Up job finished: %s" % path)

        return outputFileStoreID


def sort(file):
    """Sorts the given file."""
    with open(file) as f:
        lines = f.readlines()

    lines.sort()

    with open(file, 'w') as f:
        for line in lines:
            f.write(line)


def merge(fileHandle1, fileHandle2, outputFileHandle):
    """
    Merges together two files maintaining sorted order.

    All handles must be text-mode streams.
    """
    line2 = fileHandle2.readline()
    for line1 in fileHandle1.readlines():
        while len(line2) != 0 and line2 <= line1:
            outputFileHandle.write(line2)
            line2 = fileHandle2.readline()
        outputFileHandle.write(line1)
    while len(line2) != 0:
        outputFileHandle.write(line2)
        line2 = fileHandle2.readline()


def copySubRangeOfFile(inputFile, fileStart, fileEnd):
    """
    Copies the range (in bytes) between fileStart and fileEnd to the given
    output file handle.
    """
    with open(inputFile) as fileHandle:
        fileHandle.seek(fileStart)
        data = fileHandle.read(fileEnd - fileStart)
        assert len(data) == fileEnd - fileStart
    return data


def getMidPoint(file, fileStart, fileEnd):
    """
    Finds the point in the file to split.
    Returns an int i such that fileStart <= i < fileEnd
    """
    with open(file) as f:
        midPoint = (fileStart + fileEnd) // 2
        assert midPoint >= fileStart
        f.seek(midPoint)
        line = f.readline()
        assert len(line) >= 1
        if len(line) + midPoint < fileEnd:
            return midPoint + len(line) - 1
        f.seek(fileStart)
        line = f.readline()
        assert len(line) >= 1
        assert len(line) + fileStart <= fileEnd
    return len(line) + fileStart - 1


def makeFileToSort(fileName, lines=defaultLines, lineLen=defaultLineLen):
    with open(fileName, 'w') as f:
        for _ in range(lines):
            line = "".join(random.choice('actgACTGNXYZ') for _ in range(lineLen - 1)) + '\n'
            f.write(line)


def main(options=None):
    if not options:
        # deal with command line arguments
        parser = ArgumentParser()
        Job.Runner.addToilOptions(parser)
        parser.add_argument('--numLines', default=defaultLines, help='Number of lines in file to sort.', type=int)
        parser.add_argument('--lineLength', default=defaultLineLen, help='Length of lines in file to sort.', type=int)
        parser.add_argument("--fileToSort", help="The file you wish to sort")
        parser.add_argument("--outputFile", help="Where the sorted output will go")
        parser.add_argument("--overwriteOutput", help="Write over the output file if it already exists.", default=True)
        parser.add_argument("--N", dest="N",
                            help="The threshold below which a serial sort function is used to sort file. "
                                 "All lines must of length less than or equal to N or program will fail",
                            default=10000)
        parser.add_argument('--downCheckpoints', action='store_true',
                            help='If this option is set, the workflow will make checkpoints on its way through'
                                 'the recursive "down" part of the sort')
        parser.add_argument("--sortMemory", dest="sortMemory",
                        help="Memory for jobs that sort chunks of the file.",
                        default=None)

        parser.add_argument("--mergeMemory", dest="mergeMemory",
                        help="Memory for jobs that collate results.",
                        default=None)

        options = parser.parse_args()
    if not hasattr(options, "sortMemory") or not options.sortMemory:
        options.sortMemory = sortMemory
    if not hasattr(options, "mergeMemory") or not options.mergeMemory:
        options.mergeMemory = sortMemory

    # do some input verification
    sortedFileName = options.outputFile or "sortedFile.txt"
    if not options.overwriteOutput and os.path.exists(sortedFileName):
        print(f'Output file {sortedFileName} already exists.  '
              f'Delete it to run the sort example again or use --overwriteOutput=True')
        exit()

    fileName = options.fileToSort
    if options.fileToSort is None:
        # make the file ourselves
        fileName = 'fileToSort.txt'
        if os.path.exists(fileName):
            print(f'Sorting existing file: {fileName}')
        else:
            print(f'No sort file specified. Generating one automatically called: {fileName}.')
            makeFileToSort(fileName=fileName, lines=options.numLines, lineLen=options.lineLength)
    else:
        if not os.path.exists(options.fileToSort):
            raise RuntimeError("File to sort does not exist: %s" % options.fileToSort)

    if int(options.N) <= 0:
        raise RuntimeError("Invalid value of N: %s" % options.N)


    # Now we are ready to run
    with Toil(options) as workflow:
        sortedFileURL = 'file://' + os.path.abspath(sortedFileName)
        #raise Exception('test')

        if not workflow.options.restart:
            sortFileURL = 'file://' + os.path.abspath(fileName)
            sortFileID = workflow.importFile(sortFileURL)
            sortedFileID = workflow.start(Job.wrapJobFn(setup,
                                                        sortFileID,
                                                        int(options.N),
                                                        options.downCheckpoints,
                                                        options=options,
                                                        memory=sortMemory))
        """
        The else block is removed here to test that the job store is not
        destroyed when attempting to resume without restart().
        """
if __name__ == '__main__':
    main()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A simple user script for Toil
"""

from configargparse import ArgumentParser

from toil.common import Toil
from toil.job import Job

childMessage = "The child job is now running!"
parentMessage = "The parent job is now running!"

def hello_world(job):

    job.fileStore.log_to_leader(parentMessage)
    with open('foo_bam.txt', 'w') as handle:
        handle.write('\nThis is a triumph...\n')

    # Assign FileStoreID to a given file
    foo_bam = job.fileStore.writeGlobalFile('foo_bam.txt')

    # Spawn child
    job.addChildJobFn(hello_world_child, foo_bam, memory=100, cores=0.5, disk="3G")


def hello_world_child(job, hw):

    path = job.fileStore.readGlobalFile(hw)
    job.fileStore.log_to_leader(childMessage)
    # NOTE: path and the udpated file are stored to /tmp
    # If we want to SAVE our changes to this tmp file, we must write it out.
    with open(path) as r:
        with open('bar_bam.txt', 'w') as handle:
            for line in r.readlines():
                handle.write(line)

    # Assign FileStoreID to a given file
    # can also use:  job.updateGlobalFile() given the FileStoreID instantiation.
    job.fileStore.writeGlobalFile('bar_bam.txt')


def main():
    # Boilerplate -- startToil requires options

    parser = ArgumentParser()
    Job.Runner.addToilOptions(parser)
    options = parser.parse_args()

    # Create object that contains our FileStoreIDs

    # Launch first toil Job
    i = Job.wrapJobFn(hello_world, memory=100, cores=0.5, disk="3G")
    with Toil(options) as toil:
        toil.start(i)


if __name__ == '__main__':
    main()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import os
import threading
import time
import uuid
from typing import Optional, Set

import requests
from libcloud.compute.drivers.gce import GCEFailedNode
from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider

from toil.jobStores.googleJobStore import GoogleJobStore
from toil.lib.conversions import human2bytes
from toil.lib.compatibility import compat_bytes_recursive
from toil.provisioners import NoSuchClusterException
from toil.provisioners.abstractProvisioner import AbstractProvisioner, Shape
from toil.provisioners.node import Node

logger = logging.getLogger(__name__)
logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)

class GCEProvisioner(AbstractProvisioner):
    """
    Implements a Google Compute Engine Provisioner using libcloud.
    """

    NODE_BOTO_PATH = "/root/.boto"  # boto file path on instances
    SOURCE_IMAGE = b'projects/kinvolk-public/global/images/family/flatcar-stable'

    def __init__(self, clusterName, clusterType, zone, nodeStorage, nodeStorageOverrides, sseKey):
        self.cloud = 'gce'
        self._sseKey = sseKey

        # Call base class constructor, which will call createClusterSettings()
        # or readClusterSettings()
        super().__init__(clusterName, clusterType, zone, nodeStorage, nodeStorageOverrides)

    def supportedClusterTypes(self):
        return {'mesos'}

    def createClusterSettings(self):
        # All we need to do is read the Google credentials we need to provision
        # things
        self._readCredentials()

    def readClusterSettings(self):
        """
        Read the cluster settings from the instance, which should be the leader.
        See https://cloud.google.com/compute/docs/storing-retrieving-metadata for details about
        reading the metadata.
        """
        metadata_server = "http://metadata/computeMetadata/v1/instance/"
        metadata_flavor = {'Metadata-Flavor': 'Google'}
        zone = requests.get(metadata_server + 'zone', headers = metadata_flavor).text
        self._zone = zone.split('/')[-1]

        project_metadata_server = "http://metadata/computeMetadata/v1/project/"
        self._projectId = requests.get(project_metadata_server + 'project-id', headers = metadata_flavor).text

        # From a GCE instance, these values can be blank. Only the projectId is needed
        self._googleJson = ''
        self._clientEmail = ''

        self._tags = requests.get(metadata_server + 'description', headers = metadata_flavor).text
        tags = json.loads(self._tags)
        self.clusterName = tags['clusterName']
        self._gceDriver = self._getDriver()
        self._instanceGroup = self._gceDriver.ex_get_instancegroup(self.clusterName, zone=self._zone)

        leader = self.getLeader()
        self._leaderPrivateIP = leader.privateIP

        # The location of the Google credentials file on instances.
        self._credentialsPath = GoogleJobStore.nodeServiceAccountJson
        self._keyName = 'core' # key name leader users to communicate with works
        self._botoPath = self.NODE_BOTO_PATH # boto credentials (used if reading an AWS bucket)

        # Let the base provisioner work out how to deploy duly authorized
        # workers for this leader.
        self._setLeaderWorkerAuthentication()

    def _readCredentials(self):
        """
        Get the credentials from the file specified by GOOGLE_APPLICATION_CREDENTIALS.
        """
        self._googleJson = os.getenv('GOOGLE_APPLICATION_CREDENTIALS')
        if not self._googleJson:
            raise RuntimeError('GOOGLE_APPLICATION_CREDENTIALS not set.')
        try:
            with open(self._googleJson) as jsonFile:
                self.googleConnectionParams = json.loads(jsonFile.read())
        except:
             raise RuntimeError('GCEProvisioner: Could not parse the Google service account json file %s'
                                % self._googleJson)

        self._projectId = self.googleConnectionParams['project_id']
        self._clientEmail = self.googleConnectionParams['client_email']
        self._credentialsPath = self._googleJson
        self._clearLeaderWorkerAuthentication()  # TODO: Why are we doing this?
        self._gceDriver = self._getDriver()

    def _write_file_to_cloud(self, key: str, contents: bytes) -> str:
        raise NotImplementedError("The gceProvisioner doesn't support _write_file_to_cloud().")

    def _get_user_data_limit(self) -> int:
        # See: https://cloud.google.com/compute/docs/metadata/setting-custom-metadata#limitations
        return human2bytes('256KB')

    def launchCluster(self, leaderNodeType, leaderStorage, owner, **kwargs):
        """
        In addition to the parameters inherited from the abstractProvisioner,
        the Google launchCluster takes the following parameters:
        keyName: The key used to communicate with instances
        botoPath: Boto credentials for reading an AWS jobStore (optional).
        network: a network (optional)
        vpcSubnet: A subnet (optional).
        use_private_ip: even though a public ip exists, ignore it (optional)
        """
        if 'keyName' not in kwargs:
            raise RuntimeError("A keyPairName is required for the GCE provisioner.")
        self._keyName = kwargs['keyName']
        if 'botoPath' in kwargs:
            self._botoPath = kwargs['botoPath']
        self._vpcSubnet = kwargs.get('vpcSubnet', None)
        self._network = kwargs.get('network', None)
        self._use_private_ip = kwargs.get('use_private_ip', None)

        # Throws an error if cluster exists
        self._instanceGroup = self._gceDriver.ex_create_instancegroup(self.clusterName, self._zone)
        logger.debug('Launching leader')

        # GCE doesn't have a dictionary tags field. The tags field is just a string list.
        # Therefore, dumping tags into the description.
        tags = {'Owner': self._keyName, 'clusterName': self.clusterName}
        if 'userTags' in kwargs:
            tags.update(kwargs['userTags'])
        self._tags = json.dumps(tags)

        metadata = {'items': [{'key': 'user-data', 'value': self._getIgnitionUserData('leader')}]}
        imageType = 'flatcar-stable'
        sa_scopes = [{'scopes': ['compute', 'storage-full']}]
        disk = {}
        disk['initializeParams'] = {
            'sourceImage': self.SOURCE_IMAGE,
            'diskSizeGb': leaderStorage
        }
        disk.update({
            'boot': True,
            'autoDelete': True
        })
        name = 'l' + str(uuid.uuid4())

        leader = self._gceDriver.create_node(
            name,
            leaderNodeType,
            imageType,
            location=self._zone,
            ex_service_accounts=sa_scopes,
            ex_metadata=compat_bytes_recursive(metadata),
            ex_network=self._network,
            ex_subnetwork=self._vpcSubnet,
            ex_disks_gce_struct = [ compat_bytes_recursive(disk) ],
            description=self._tags,
            ex_preemptible=False
        )

        self._instanceGroup.add_instances([leader])
        self._leaderPrivateIP = leader.private_ips[0]  # needed if adding workers
        # self.subnetID = leader.subnet_id  # TODO: get subnetID

        # Wait for the appliance to start and inject credentials.
        leaderNode = Node(publicIP=leader.public_ips[0], privateIP=leader.private_ips[0],
                          name=leader.name, launchTime=leader.created_at, nodeType=leader.size,
                          preemptible=False, tags=self._tags, use_private_ip=self._use_private_ip)
        leaderNode.waitForNode('toil_leader', keyName=self._keyName)
        leaderNode.copySshKeys(self._keyName)
        leaderNode.injectFile(self._credentialsPath, GoogleJobStore.nodeServiceAccountJson, 'toil_leader')
        if self._botoPath:
            leaderNode.injectFile(self._botoPath, self.NODE_BOTO_PATH, 'toil_leader')
        # Download credentials
        self._setLeaderWorkerAuthentication(leaderNode)

        logger.debug('Launched leader')

    def getNodeShape(self, instance_type: str, preemptible=False) -> Shape:
        # TODO: read this value only once
        sizes = self._gceDriver.list_sizes(location=self._zone)
        sizes = [x for x in sizes if x.name == instance_type]
        assert len(sizes) == 1
        instanceType = sizes[0]

        disk = 0  # instanceType.disks * instanceType.disk_capacity * 2 ** 30
        if disk == 0:
            # This is an EBS-backed instance. We will use the root
            # volume, so add the amount of EBS storage requested forhe root volume
            disk = self._nodeStorageOverrides.get(instance_type, self._nodeStorage) * 2 ** 30

        # Ram is in M.
        # Underestimate memory by 100M to prevent autoscaler from disagreeing with
        # mesos about whether a job can run on a particular node type
        memory = (instanceType.ram/1000 - 0.1) * 2 ** 30
        return Shape(wallTime=60 * 60,
                     memory=memory,
                     cores=instanceType.extra['guestCpus'],
                     disk=disk,
                     preemptible=preemptible)

    @staticmethod
    def retryPredicate(e):
        """ Not used by GCE """
        return False

    def destroyCluster(self) -> None:
        """
        Try a few times to terminate all of the instances in the group.
        """
        logger.debug("Destroying cluster %s" % self.clusterName)
        instancesToTerminate = self._getNodesInCluster()
        attempts = 0
        while instancesToTerminate and attempts < 3:
            self._terminateInstances(instances=instancesToTerminate)
            instancesToTerminate = self._getNodesInCluster()
            attempts += 1

        # remove group
        instanceGroup = self._gceDriver.ex_get_instancegroup(self.clusterName, zone=self._zone)
        instanceGroup.destroy()

    def terminateNodes(self, nodes):
        if nodes:
            nodeNames = [n.name for n in nodes]
            instances = self._getNodesInCluster()
            instancesToKill = [i for i in instances if i.name in nodeNames]
            self._terminateInstances(instancesToKill)

    def addNodes(self, nodeTypes: Set[str], numNodes, preemptible, spotBid=None) -> int:
        assert self._leaderPrivateIP

        # We don't support any balancing here so just pick one of the
        # equivalent node types
        node_type = next(iter(nodeTypes))

        # If keys are rsynced, then the mesos-slave needs to be started after the keys have been
        # transferred. The waitForKey.sh script loops on the new VM until it finds the keyPath file, then it starts the
        # mesos-slave. If there are multiple keys to be transferred, then the last one to be transferred must be
        # set to keyPath.
        keyPath = None
        botoExists = False
        if self._botoPath is not None and os.path.exists(self._botoPath):
            keyPath = self.NODE_BOTO_PATH
            botoExists = True
        elif self._sseKey:
            keyPath = self._sseKey

        if not preemptible:
            logger.debug('Launching %s non-preemptible nodes', numNodes)
        else:
            logger.debug('Launching %s preemptible nodes', numNodes)

        # kwargs["subnet_id"] = self.subnetID if self.subnetID else self._getClusterInstance(self.instanceMetaData).subnet_id
        userData = self._getIgnitionUserData('worker', keyPath, preemptible)
        metadata = {'items': [{'key': 'user-data', 'value': userData}]}
        imageType = 'flatcar-stable'
        sa_scopes = [{'scopes': ['compute', 'storage-full']}]
        disk = {}
        disk['initializeParams'] = {
            'sourceImage': self.SOURCE_IMAGE,
            'diskSizeGb': self._nodeStorageOverrides.get(node_type, self._nodeStorage) }
        disk.update({
            'boot': True,
            'autoDelete': True
        })

        # TODO:
        #  - bug in gce.py for ex_create_multiple_nodes (erroneously, doesn't allow image and disk to specified)
        #  - ex_create_multiple_nodes is limited to 1000 nodes
        #    - use a different function
        #    - or write a loop over the rest of this function, with 1K nodes max on each iteration
        retries = 0
        workersCreated = 0
        # Try a few times to create the requested number of workers
        while numNodes-workersCreated > 0 and retries < 3:
            instancesLaunched = self.ex_create_multiple_nodes(
                                    '', node_type, imageType, numNodes-workersCreated,
                                    location=self._zone,
                                    ex_service_accounts=sa_scopes,
                                    ex_metadata=metadata,
                                    ex_disks_gce_struct=[disk],
                                    description=self._tags,
                                    ex_preemptible=preemptible
                                    )
            failedWorkers = []
            for instance in instancesLaunched:
                if isinstance(instance, GCEFailedNode):
                    logger.error("Worker failed to launch with code %s. Error message: %s"
                                 % (instance.code, instance.error))
                    continue

                node = Node(publicIP=instance.public_ips[0], privateIP=instance.private_ips[0],
                            name=instance.name, launchTime=instance.created_at, nodeType=instance.size,
                            preemptible=False, tags=self._tags, use_private_ip=self._use_private_ip)  # FIXME: what should tags be set to?

                try:
                    self._injectWorkerFiles(node, botoExists)
                    logger.debug("Created worker %s" % node.publicIP)
                    self._instanceGroup.add_instances([instance])
                    workersCreated += 1
                except Exception as e:
                    logger.error(f"Failed to configure worker {node.name}. Error message: {e}")
                    failedWorkers.append(instance)
            if failedWorkers:
                logger.error("Terminating %d failed workers" % len(failedWorkers))
                self._terminateInstances(failedWorkers)
            retries += 1

        logger.debug('Launched %d new instance(s)', numNodes)
        if numNodes != workersCreated:
            logger.error("Failed to launch %d worker(s)", numNodes-workersCreated)
        return workersCreated

    def getProvisionedWorkers(self, instance_type: Optional[str] = None, preemptible: Optional[bool] = None):
        assert self._leaderPrivateIP
        entireCluster = self._getNodesInCluster(instance_type=instance_type)
        logger.debug('All nodes in cluster: %s', entireCluster)
        workerInstances = []
        for instance in entireCluster:
            if preemptible is not None:
                scheduling = instance.extra.get('scheduling')
                # If this field is not found in the extra meta-data, assume the node is not preemptible.
                if scheduling and scheduling.get('preemptible', False) != preemptible:
                    continue
            isWorker = True
            for ip in instance.private_ips:
                if ip == self._leaderPrivateIP:
                    isWorker = False
                    break  # don't include the leader
            if isWorker and instance.state == 'running':
                workerInstances.append(instance)

        logger.debug('All workers found in cluster: %s', workerInstances)
        return [Node(publicIP=i.public_ips[0], privateIP=i.private_ips[0],
                     name=i.name, launchTime=i.created_at, nodeType=i.size,
                     preemptible=i.extra.get('scheduling', {}).get('preemptible', False),
                     tags=None, use_private_ip=self._use_private_ip)
                for i in workerInstances]

    def getLeader(self):
        instances = self._getNodesInCluster()
        instances.sort(key=lambda x: x.created_at)
        try:
            leader = instances[0]  # assume leader was launched first
        except IndexError:
            raise NoSuchClusterException(self.clusterName)
        return Node(publicIP=leader.public_ips[0], privateIP=leader.private_ips[0],
                    name=leader.name, launchTime=leader.created_at, nodeType=leader.size,
                    preemptible=False, tags=None, use_private_ip=self._use_private_ip)

    def _injectWorkerFiles(self, node, botoExists):
        """
        Set up the credentials on the worker.
        """
        node.waitForNode('toil_worker', keyName=self._keyName)
        node.copySshKeys(self._keyName)
        node.injectFile(self._credentialsPath, GoogleJobStore.nodeServiceAccountJson, 'toil_worker')
        if self._sseKey:
            node.injectFile(self._sseKey, self._sseKey, 'toil_worker')
        if botoExists:
            node.injectFile(self._botoPath, self.NODE_BOTO_PATH, 'toil_worker')

    def _getNodesInCluster(self, instance_type: Optional[str] = None):
        instanceGroup = self._gceDriver.ex_get_instancegroup(self.clusterName, zone=self._zone)
        instances = instanceGroup.list_instances()
        if instance_type:
            instances = [instance for instance in instances if instance.size == instance_type]
        return instances

    def _getDriver(self):
        """  Connect to GCE """
        driverCls = get_driver(Provider.GCE)
        return driverCls(self._clientEmail,
                         self._googleJson,
                         project=self._projectId,
                         datacenter=self._zone)

    def _terminateInstances(self, instances):
        def worker(driver, instance):
            logger.debug('Terminating instance: %s', instance.name)
            driver.destroy_node(instance)

        threads = []
        for instance in instances:
            t = threading.Thread(target=worker, args=(self._gceDriver,instance))
            threads.append(t)
            t.start()

        logger.debug('... Waiting for instance(s) to shut down...')
        for t in threads:
            t.join()

    # MONKEY PATCH - This function was copied form libcloud to fix a bug.
    DEFAULT_TASK_COMPLETION_TIMEOUT = 180

    def ex_create_multiple_nodes(
            self, base_name, size, image, number, location=None,
            ex_network='default', ex_subnetwork=None, ex_tags=None,
            ex_metadata=None, ignore_errors=True, use_existing_disk=True,
            poll_interval=2, external_ip='ephemeral',
            ex_disk_type='pd-standard', ex_disk_auto_delete=True,
            ex_service_accounts=None, timeout=DEFAULT_TASK_COMPLETION_TIMEOUT,
            description=None, ex_can_ip_forward=None, ex_disks_gce_struct=None,
            ex_nic_gce_struct=None, ex_on_host_maintenance=None,
            ex_automatic_restart=None, ex_image_family=None,
            ex_preemptible=None):
        """
         Monkey patch to gce.py in libcloud to allow disk and images to be specified.
         Also changed name to a uuid below.
         The prefix 'wp' identifies preemptible nodes and 'wn' non-preemptible nodes.
        """
        # if image and ex_disks_gce_struct:
        #    raise ValueError("Cannot specify both 'image' and "
        #                     "'ex_disks_gce_struct'.")

        driver = self._getDriver()
        if image and ex_image_family:
            raise ValueError("Cannot specify both 'image' and "
                             "'ex_image_family'")

        location = location or driver.zone
        if not hasattr(location, 'name'):
            location = driver.ex_get_zone(location)
        if not hasattr(size, 'name'):
            size = driver.ex_get_size(size, location)
        if not hasattr(ex_network, 'name'):
            ex_network = driver.ex_get_network(ex_network)
        if ex_subnetwork and not hasattr(ex_subnetwork, 'name'):
            ex_subnetwork = \
                driver.ex_get_subnetwork(ex_subnetwork,
                                         region=driver._get_region_from_zone(location))
        if ex_image_family:
            image = driver.ex_get_image_from_family(ex_image_family)
        if image and not hasattr(image, 'name'):
            image = driver.ex_get_image(image)
        if not hasattr(ex_disk_type, 'name'):
            ex_disk_type = driver.ex_get_disktype(ex_disk_type, zone=location)

        node_attrs = {'size': size,
                      'image': image,
                      'location': location,
                      'network': ex_network,
                      'subnetwork': ex_subnetwork,
                      'tags': ex_tags,
                      'metadata': ex_metadata,
                      'ignore_errors': ignore_errors,
                      'use_existing_disk': use_existing_disk,
                      'external_ip': external_ip,
                      'ex_disk_type': ex_disk_type,
                      'ex_disk_auto_delete': ex_disk_auto_delete,
                      'ex_service_accounts': ex_service_accounts,
                      'description': description,
                      'ex_can_ip_forward': ex_can_ip_forward,
                      'ex_disks_gce_struct': ex_disks_gce_struct,
                      'ex_nic_gce_struct': ex_nic_gce_struct,
                      'ex_on_host_maintenance': ex_on_host_maintenance,
                      'ex_automatic_restart': ex_automatic_restart,
                      'ex_preemptible': ex_preemptible}
        # List for holding the status information for disk/node creation.
        status_list = []

        for i in range(number):
            name = 'wp' if ex_preemptible else 'wn'
            name += str(uuid.uuid4())  # '%s-%03d' % (base_name, i)
            status = {'name': name, 'node_response': None, 'node': None}
            status_list.append(status)

        start_time = time.time()
        complete = False
        while not complete:
            if time.time() - start_time >= timeout:
                raise Exception("Timeout (%s sec) while waiting for multiple "
                                "instances")
            complete = True
            time.sleep(poll_interval)
            for status in status_list:
                # Create the node or check status if already in progress.
                if not status['node']:
                    if not status['node_response']:
                        driver._multi_create_node(status, node_attrs)
                    else:
                        driver._multi_check_node(status, node_attrs)
                # If any of the nodes have not been created (or failed) we are
                # not done yet.
                if not status['node']:
                    complete = False

        # Return list of nodes
        node_list = []
        for status in status_list:
            node_list.append(status['node'])
        return node_list



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import time
from queue import Empty, Queue
from threading import Event, Thread
from typing import Iterable, Optional, Set

from toil.job import ServiceJobDescription
from toil.jobStores.abstractJobStore import AbstractJobStore
from toil.lib.throttle import LocalThrottle, throttle
from toil.toilState import ToilState

logger = logging.getLogger(__name__)


class ServiceManager:
    """Manages the scheduling of services."""

    def __init__(self, job_store: AbstractJobStore, toil_state: ToilState) -> None:
        logger.debug("Initializing service manager")
        self.__job_store = job_store

        self.__toil_state = toil_state

        # We call the jobs that have services they need "client" jobs.

        # These are all the client jobs that are waiting for their services to
        # start.
        self.__waiting_clients: Set[str] = set()

        # This is used to terminate the thread associated with the service
        # manager
        self.__terminate = Event()

        # This is the input queue of jobs that have services that need to be started
        self.__clients_in: Queue[str] = Queue()

        # This is the output queue of jobs that have services that
        # are already started
        self.__clients_out: Queue[str] = Queue()

        # This is the output queue of jobs that have services that are unable
        # to start
        self.__failed_clients_out: Queue[str] = Queue()

        # This is the queue of services for the batch system to start
        self.__services_out: Queue[str] = Queue()

        # The number of jobs the service manager is scheduling
        self.__service_manager_jobs = 0

        # Set up the service-managing thread.
        self.__service_starter = Thread(target=self.__start_services, daemon=True)

    def services_are_starting(self, job_id: str) -> bool:
        """
        Check if services are being started.

        :return: True if the services for the given job are currently being started, and False otherwise.
        """
        return job_id in self.__waiting_clients

    def get_job_count(self) -> int:
        """
        Get the total number of jobs we are working on.

        (services and their parent non-service jobs)
        """
        return self.__service_manager_jobs

    def start(self) -> None:
        """Start the service scheduling thread."""
        self.__service_starter.start()

    def put_client(self, client_id: str) -> None:
        """
        Schedule the services of a job asynchronously.

        When the job's services are running the ID for the job will
        be returned by toil.leader.ServiceManager.get_ready_client.

        :param client_id: ID of job with services to schedule.
        """
        # Go get the client's description, which includes the services it needs.
        client = self.__toil_state.get_job(client_id)

        logger.debug("Service manager queueing %s as client", client)

        # Add job to set being processed by the service manager
        self.__waiting_clients.add(client_id)

        # Add number of jobs managed by ServiceManager
        self.__service_manager_jobs += (
            len(client.services) + 1
        )  # The plus one accounts for the root job

        # Asynchronously schedule the services
        self.__clients_in.put(client_id)

    def get_ready_client(self, maxWait: float) -> Optional[str]:
        """
        Fetch a ready client, waiting as needed.

        :param float maxWait: Time in seconds to wait to get a JobDescription before returning
        :return: the ID of a client whose services are running, or None if no
                 such job is available.
        """
        try:
            client_id = self.__clients_out.get(timeout=maxWait)
            self.__waiting_clients.remove(client_id)
            if self.__service_manager_jobs < 0:
                raise RuntimeError("The number of jobs scheduled by the service manager cannot be negative.")
            self.__service_manager_jobs -= 1
            return client_id
        except Empty:
            return None

    def get_unservable_client(self, maxWait: float) -> Optional[str]:
        """
        Fetch a client whos services failed to start.

        :param float maxWait: Time in seconds to wait to get a JobDescription before returning
        :return: the ID of a client whose services failed to start, or None if
                 no such job is available.
        """
        try:
            client_id = self.__failed_clients_out.get(timeout=maxWait)
            self.__waiting_clients.remove(client_id)
            if self.__service_manager_jobs < 0:
                raise RuntimeError("The number of jobs scheduled by the service manager cannot be negative.")
            self.__service_manager_jobs -= 1
            return client_id
        except Empty:
            return None

    def get_startable_service(self, maxWait: float) -> Optional[str]:
        """
        Fetch a service job that is ready to start.

        :param maxWait: Time in seconds to wait to get a job before returning.
        :return: the ID of a service job that the leader can start, or None if no such job exists.
        """
        try:
            service_id = self.__services_out.get(timeout=maxWait)
            if self.__service_manager_jobs < 0:
                raise RuntimeError("The number of jobs scheduled by the service manager cannot be negative.")
            self.__service_manager_jobs -= 1
            return service_id
        except Empty:
            return None

    def _get_service_job(self, service_id: str) -> ServiceJobDescription:
        service = self.__toil_state.get_job(service_id)
        if not isinstance(service, ServiceJobDescription):
            raise Exception(
                f"Expected ServiceJobDescription, got '{type(service)}': {service}."
            )
        return service

    def kill_services(self, service_ids: Iterable[str], error: bool = False) -> None:
        """
        Stop all the given service jobs.

        :param services: Service jobStoreIDs to kill
        :param error: Whether to signal that the service failed with an error when stopping it.
        """
        for service_id in service_ids:
            # Get the job description, which knows about the flag files.
            service = self._get_service_job(service_id)
            if service.errorJobStoreID is None or service.terminateJobStoreID is None:
                raise Exception("Must be a registered ServiceJobDescription")
            if error:
                self.__job_store.delete_file(service.errorJobStoreID)
            self.__job_store.delete_file(service.terminateJobStoreID)

    def is_active(self, service_id: str) -> bool:
        """
        Return true if the service job has not been told to terminate.

        :param service_id: Service to check on
        """
        service = self._get_service_job(service_id)
        if service.terminateJobStoreID is None:
            raise Exception("Must be a registered ServiceJobDescription")
        return self.__job_store.file_exists(service.terminateJobStoreID)

    def is_running(self, service_id: str) -> bool:
        """
        Return true if the service job has started and is active.

        :param service: Service to check on
        """
        service = self._get_service_job(service_id)
        if service.startJobStoreID is None:
            raise Exception("Must be a registered ServiceJobDescription")
        return (
            not self.__job_store.file_exists(service.startJobStoreID)
        ) and self.is_active(service_id)

    def check(self) -> None:
        """
        Check on the service manager thread.

        :raise RuntimeError: If the underlying thread has quit.
        """
        if not self.__service_starter.is_alive():
            raise RuntimeError("Service manager has quit")

    def shutdown(self) -> None:
        """
        Terminate worker threads cleanly; starting and killing all service threads.

        Will block until all services are started and blocked.
        """
        logger.debug('Waiting for service manager thread to finish ...')
        start_time = time.time()
        self.__terminate.set()
        self.__service_starter.join()
        # Kill any services still running to avoid deadlock
        for services in list(self.__toil_state.servicesIssued.values()):
            self.kill_services(services, error=True)
        logger.debug(
            "... finished shutting down the service manager. Took %s seconds",
            time.time() - start_time,
        )

    def __start_services(self) -> None:
        """Thread used to schedule services."""
        # Keep the user informed, but not too informed, as services start up
        log_limiter = LocalThrottle(60)

        # These are all keyed by ID
        starting_services = set()
        remaining_services_by_client = {}
        service_to_client = {}
        clients_with_failed_services = set()
        while True:
            with throttle(1.0):
                if self.__terminate.is_set():
                    logger.debug('Received signal to quit starting services.')
                    break
                try:
                    client_id = self.__clients_in.get_nowait()
                    client = self.__toil_state.get_job(client_id)
                    host_id_batches = list(client.serviceHostIDsInBatches())
                    logger.debug("Service manager processing client %s with %d batches of services", client, len(host_id_batches))
                    if len(host_id_batches) > 1:
                        # Have to fall back to the old blocking behavior to
                        # ensure entire service "groups" are issued as a whole.
                        self.__start_batches_blocking(client_id)
                        continue
                    # Found a new job that needs to schedule its services.
                    for batch in host_id_batches:
                        # There should be just one batch so we can do it here.
                        remaining_services_by_client[client_id] = len(batch)
                        for service_id in batch:
                            # Load up the service object.
                            service_job_desc = self._get_service_job(service_id)
                            # Remember the parent job
                            service_to_client[service_id] = client_id
                            # We should now start to monitor this service to see if
                            # it has started yet.
                            starting_services.add(service_id)
                            # Send the service JobDescription off to be started
                            logger.debug(
                                "Service manager is starting service job: %s, start ID: %s",
                                service_job_desc,
                                service_job_desc.startJobStoreID,
                            )
                            self.__services_out.put(service_id)
                except Empty:
                    # No new jobs that need services scheduled.
                    pass

                pending_service_count = len(starting_services)
                if pending_service_count > 0 and log_limiter.throttle(False):
                    logger.debug('%d services are starting...', pending_service_count)

                for service_id in list(starting_services):
                    service_job_desc = self._get_service_job(service_id)
                    if (
                        service_job_desc.startJobStoreID is None
                        or service_job_desc.errorJobStoreID is None
                    ):
                        raise Exception("Must be a registered ServiceJobDescription")
                    if not self.__job_store.file_exists(service_job_desc.startJobStoreID):
                        # Service has started (or failed)
                        logger.debug(
                            "Service %s has removed %s and is therefore started",
                            service_job_desc,
                            service_job_desc.startJobStoreID,
                        )
                        starting_services.remove(service_id)
                        client_id = service_to_client[service_id]
                        remaining_services_by_client[client_id] -= 1
                        if remaining_services_by_client[client_id] < 0:
                            raise RuntimeError("The number of remaining services cannot be negative.")
                        del service_to_client[service_id]
                        if not self.__job_store.file_exists(service_job_desc.errorJobStoreID):
                            logger.error(
                                "Service %s has immediately failed before it could be used",
                                service_job_desc,
                            )
                            # It probably hasn't fileld in the promise that the
                            # job that uses the service needs.
                            clients_with_failed_services.add(client_id)

                # Find if any clients have had *all* their services started.
                ready_clients = set()
                for client_id, remainingServices in remaining_services_by_client.items():
                    if remainingServices == 0:
                        if client_id in clients_with_failed_services:
                            logger.error('Job %s has had all its services try to start, but at least one failed', self.__toil_state.get_job(client_id))
                            self.__failed_clients_out.put(client_id)
                        else:
                            logger.debug('Job %s has all its services started', self.__toil_state.get_job(client_id))
                            self.__clients_out.put(client_id)
                        ready_clients.add(client_id)
                for client_id in ready_clients:
                    del remaining_services_by_client[client_id]

    def __start_batches_blocking(self, client_id: str) -> None:
        """
        Wait until all the services for the given job are started.

        (Starting them in batches that are all issued together)
        """
        # Keep the user informed, but not too informed, as services start up
        log_limiter = LocalThrottle(60)

        # Start the service jobs in batches, waiting for each batch
        # to become established before starting the next batch
        for service_job_list in self.__toil_state.get_job(client_id).serviceHostIDsInBatches():
            # When we get the job descriptions we store them here to go over them again.
            wait_on = []
            for service_id in service_job_list:
                # Find the service object.
                service_job_desc = self._get_service_job(service_id)
                if (
                    service_job_desc.startJobStoreID is None
                    or service_job_desc.jobStoreID is None
                ):
                    raise Exception("Must be a registered ServiceJobDescription")
                logger.debug(
                    "Service manager is starting service job: %s, start ID: %s",
                    service_job_desc,
                    service_job_desc.startJobStoreID,
                )
                if not self.__job_store.file_exists(service_job_desc.startJobStoreID):
                    raise RuntimeError(f"Service manager attempted to start service {service_job_desc} that has already started")
                if not self.__toil_state.job_exists(str(service_job_desc.jobStoreID)):
                    raise RuntimeError(f"Service manager attempted to start service {service_job_desc} that is not in the job store")
                # At this point the terminateJobStoreID and errorJobStoreID
                # could have been deleted, since the service can be killed at
                # any time! So we can't assert their presence here.
                self.__services_out.put(service_id)
                # Save for the waiting loop
                wait_on.append(service_job_desc)

            # Wait until all the services of the batch are running
            for service_id in service_job_list:
                # Find the service object.
                service_job_desc = self._get_service_job(service_id)
                if service_job_desc.startJobStoreID is None:
                    raise Exception("Must be a registered ServiceJobDescription")
                while self.__job_store.file_exists(service_job_desc.startJobStoreID):
                    # Sleep to avoid thrashing
                    time.sleep(1.0)

                    if log_limiter.throttle(False):
                        logger.info('Service %s is starting...', service_job_desc)

                    # Check if the thread should quit
                    if self.__terminate.is_set():
                        return

                    if not self.__toil_state.job_exists(
                        str(service_job_desc.jobStoreID)
                    ) and self.__job_store.file_exists(
                        service_job_desc.startJobStoreID
                    ):
                        # The service job has gone away but the service never flipped its start flag.
                        # That's not what the worker is supposed to do when running a service at all.
                        logger.error('Service %s has completed and been removed without ever starting', service_job_desc)
                        # Stop everything.
                        raise RuntimeError(f"Service {service_job_desc} is in an inconsistent state")

                # We don't bail out early here.

                # We need to try and fail to start *all* the services, so they
                # *all* come back to the leader as expected, or the leader will get
                # stuck waiting to hear about a later dependent service failing. So
                # we have to *try* to start all the services, even if the services
                # they depend on failed. They should already have been killed,
                # though, so they should stop immediately when we run them. TODO:
                # this is a bad design!


        # Add the JobDescription to the output queue of jobs whose services have been started
        self.__clients_out.put(client_id)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import Dict, List, Optional

from toil.batchSystems.abstractBatchSystem import (BatchSystemSupport,
                                                   UpdatedBatchJobInfo)
from toil.batchSystems.singleMachine import SingleMachineBatchSystem
from toil.common import Config
from toil.job import JobDescription
from toil.lib.threading import cpu_count

logger = logging.getLogger(__name__)


class BatchSystemLocalSupport(BatchSystemSupport):
    """Adds a local queue for helper jobs, useful for CWL & others."""

    def __init__(self, config: Config, maxCores: float, maxMemory: int, maxDisk: int) -> None:
        super().__init__(config, maxCores, maxMemory, maxDisk)
        max_local_jobs = config.max_local_jobs if config.max_local_jobs is not None else cpu_count()
        self.localBatch: SingleMachineBatchSystem = SingleMachineBatchSystem(
            config, maxCores, maxMemory, maxDisk, max_jobs=max_local_jobs
        )

    def handleLocalJob(self, jobDesc: JobDescription) -> Optional[int]:
        """
        To be called by issueBatchJobs.

        Returns the jobID if the jobDesc has been submitted to the local queue,
        otherwise returns None
        """
        if (not self.config.run_local_jobs_on_workers
                and jobDesc.local):
            # Since singleMachine.py doesn't typecheck yet and MyPy is ignoring
            # it, it will raise errors here unless we add type annotations to
            # everything we get back from it. The easiest way to do that seems
            # to be to put it in a variable with a type annotation on it. That
            # somehow doesn't error whereas just returning the value complains
            # we're returning an Any. TODO: When singleMachine.py typechecks,
            # remove all these extra variables.
            local_id: int = self.localBatch.issueBatchJob(jobDesc)
            return local_id
        else:
            return None

    def killLocalJobs(self, jobIDs: List[int]) -> None:
        """
        Will kill all local jobs that match the provided jobIDs.

        To be called by killBatchJobs.
        """
        self.localBatch.killBatchJobs(jobIDs)

    def getIssuedLocalJobIDs(self) -> List[int]:
        """To be called by getIssuedBatchJobIDs."""
        local_ids: List[int] = self.localBatch.getIssuedBatchJobIDs()
        return local_ids

    def getRunningLocalJobIDs(self) -> Dict[int, float]:
        """To be called by getRunningBatchJobIDs()."""
        local_running: Dict[int, float] = self.localBatch.getRunningBatchJobIDs()
        return local_running

    def getUpdatedLocalJob(self, maxWait: int) -> Optional[UpdatedBatchJobInfo]:
        """To be called by getUpdatedBatchJob()."""
        return self.localBatch.getUpdatedBatchJob(maxWait)

    def getNextJobID(self) -> int:
        """Must be used to get job IDs so that the local and batch jobs do not conflict."""
        # TODO: This reaches deep into SingleMachineBatchSystem when it probably shouldn't
        with self.localBatch.jobIndexLock:
            jobID: int = self.localBatch.jobIndex
            self.localBatch.jobIndex += 1
        return jobID

    def shutdownLocal(self) -> None:
        """To be called from shutdown()."""
        self.localBatch.shutdown()



import json
import os
import shutil
import subprocess
import unittest
from uuid import uuid4
from typing import Optional

from unittest.mock import patch
from typing import Any, Dict, List, Set

import pytest

from toil.provisioners import cluster_factory
from toil.test import (ToilTest,
                       needs_docker_cuda,
                       needs_google_storage,
                       needs_singularity_or_docker,
                       slow, integrative)
from toil.test.provisioners.clusterTest import AbstractClusterTest
from toil.version import exactPython
from toil.wdl.wdltoil import WDLSectionJob, WDLWorkflowGraph


class BaseWDLTest(ToilTest):
    """Base test class for WDL tests."""

    def setUp(self) -> None:
        """Runs anew before each test to create farm fresh temp dirs."""
        self.output_dir = os.path.join('/tmp/', 'toil-wdl-test-' + str(uuid4()))
        os.makedirs(self.output_dir)

    def tearDown(self) -> None:
        if os.path.exists(self.output_dir):
            shutil.rmtree(self.output_dir)


class WDLConformanceTests(BaseWDLTest):
    """
    WDL conformance tests for Toil.
    """
    wdl_dir = "wdl-conformance-tests"

    @classmethod
    def setUpClass(cls) -> None:

        url = "https://github.com/DataBiosphere/wdl-conformance-tests.git"
        commit = "032fb99a1458d456b6d5f17d27928469ec1a1c68"

        p = subprocess.Popen(
            f"git clone {url} {cls.wdl_dir} && cd {cls.wdl_dir} && git checkout {commit}",
            shell=True,
        )

        p.communicate()

        if p.returncode > 0:
            raise RuntimeError

        os.chdir(cls.wdl_dir)

        cls.base_command = [exactPython, "run.py", "--runner", "toil-wdl-runner"]

    # estimated running time: 2 minutes
    @slow
    def test_conformance_tests_v10(self):
        tests_to_run = "0,1,5-7,9-15,17,22-24,26,28-30,32-40,53,57-59,62,67-69"
        p = subprocess.run(self.base_command + ["-v", "1.0", "-n", tests_to_run], capture_output=True)

        if p.returncode != 0:
            print(p.stdout.decode('utf-8', errors='replace'))

        p.check_returncode()

    # estimated running time: 2 minutes
    @slow
    def test_conformance_tests_v11(self):
        tests_to_run = "2-11,13-15,17-20,22-24,26,29,30,32-40,53,57-59,62,67-69"
        p = subprocess.run(self.base_command + ["-v", "1.1", "-n", tests_to_run], capture_output=True)

        if p.returncode != 0:
            print(p.stdout.decode('utf-8', errors='replace'))

        p.check_returncode()

    @classmethod
    def tearDownClass(cls) -> None:
        upper_dir = os.path.dirname(os.getcwd())
        os.chdir(upper_dir)
        shutil.rmtree("wdl-conformance-tests")


class WDLTests(BaseWDLTest):
    """Tests for Toil's MiniWDL-based implementation."""

    @classmethod
    def setUpClass(cls) -> None:
        """Runs once for all tests."""
        cls.base_command = [exactPython, '-m', 'toil.wdl.wdltoil']

    # We inherit a testMD5sum but it is going to need Singularity or Docker
    # now. And also needs to have a WDL 1.0+ WDL file. So we replace it.
    @needs_singularity_or_docker
    def test_MD5sum(self):
        """Test if Toil produces the same outputs as known good outputs for WDL's
        GATK tutorial #1."""
        wdl = os.path.abspath('src/toil/test/wdl/md5sum/md5sum.1.0.wdl')
        json_file = os.path.abspath('src/toil/test/wdl/md5sum/md5sum.json')

        result_json = subprocess.check_output(
            self.base_command + [wdl, json_file, '-o', self.output_dir, '--logDebug', '--retryCount=0'])
        result = json.loads(result_json)

        assert 'ga4ghMd5.value' in result
        assert isinstance(result['ga4ghMd5.value'], str)
        assert os.path.exists(result['ga4ghMd5.value'])
        assert os.path.basename(result['ga4ghMd5.value']) == 'md5sum.txt'

    @needs_singularity_or_docker
    def test_miniwdl_self_test(self, extra_args: Optional[List[str]] = None) -> None:
        """Test if the MiniWDL self test runs and produces the expected output."""
        wdl_file = os.path.abspath('src/toil/test/wdl/miniwdl_self_test/self_test.wdl')
        json_file = os.path.abspath('src/toil/test/wdl/miniwdl_self_test/inputs.json')

        result_json = subprocess.check_output(
            self.base_command + [wdl_file, json_file, '--logDebug', '-o', self.output_dir, '--outputDialect',
                                 'miniwdl'] + (extra_args or []))
        result = json.loads(result_json)

        # Expect MiniWDL-style output with a designated "dir"

        assert 'dir' in result
        assert isinstance(result['dir'], str)
        out_dir = result['dir']

        assert 'outputs' in result
        assert isinstance(result['outputs'], dict)
        outputs = result['outputs']

        assert 'hello_caller.message_files' in outputs
        assert isinstance(outputs['hello_caller.message_files'], list)
        assert len(outputs['hello_caller.message_files']) == 2
        for item in outputs['hello_caller.message_files']:
            # All the files should be strings in the "out" direcotry
            assert isinstance(item, str)
            assert item.startswith(out_dir)

        assert 'hello_caller.messages' in outputs
        assert outputs['hello_caller.messages'] == ["Hello, Alyssa P. Hacker!", "Hello, Ben Bitdiddle!"]

    @needs_singularity_or_docker
    def test_miniwdl_self_test_by_reference(self) -> None:
        """
        Test if the MiniWDL self test works when passing input files by URL reference.
        """
        self.test_miniwdl_self_test(extra_args=["--referenceInputs=True"])

    @slow
    @needs_docker_cuda
    def test_giraffe_deepvariant(self):
        """Test if Giraffe and GPU DeepVariant run. This could take 25 minutes."""
        # TODO: enable test if nvidia-container-runtime and Singularity are installed but Docker isn't.

        json_dir = self._createTempDir()
        base_uri = 'https://raw.githubusercontent.com/vgteam/vg_wdl/65dd739aae765f5c4dedd14f2e42d5a263f9267a'

        wdl_file = f"{base_uri}/workflows/giraffe_and_deepvariant.wdl"
        json_file = os.path.abspath(os.path.join(json_dir, 'inputs.json'))
        with open(json_file, 'w') as fp:
            # Write some inputs. We need to override the example inputs to use a GPU container, but that means we need absolute input URLs.
            json.dump(fp, {
                "GiraffeDeepVariant.INPUT_READ_FILE_1": f"{base_uri}/tests/small_sim_graph/reads_1.fastq.gz",
                "GiraffeDeepVariant.INPUT_READ_FILE_2": f"{base_uri}/tests/small_sim_graph/reads_2.fastq.gz",
                "GiraffeDeepVariant.XG_FILE": f"{base_uri}/tests/small_sim_graph/graph.xg",
                "GiraffeDeepVariant.SAMPLE_NAME": "s0",
                "GiraffeDeepVariant.GBWT_FILE": f"{base_uri}/tests/small_sim_graph/graph.gbwt",
                "GiraffeDeepVariant.GGBWT_FILE": f"{base_uri}/tests/small_sim_graph/graph.gg",
                "GiraffeDeepVariant.MIN_FILE": f"{base_uri}/tests/small_sim_graph/graph.min",
                "GiraffeDeepVariant.DIST_FILE": f"{base_uri}/tests/small_sim_graph/graph.dist",
                "GiraffeDeepVariant.OUTPUT_GAF": True,
                "GiraffeDeepVariant.runDeepVariantCallVariants.in_dv_gpu_container": "google/deepvariant:1.3.0-gpu"
            })

        result_json = subprocess.check_output(
            self.base_command + [wdl_file, json_file, '-o', self.output_dir, '--outputDialect', 'miniwdl'])
        result = json.loads(result_json)

        # Expect MiniWDL-style output with a designated "dir"
        assert 'dir' in result
        assert isinstance(result['dir'], str)
        out_dir = result['dir']

        assert 'outputs' in result
        assert isinstance(result['outputs'], dict)
        outputs = result['outputs']

        # Expect a VCF file to have been written
        assert 'GiraffeDeepVariant.output_vcf' in outputs
        assert isinstance(outputs['GiraffeDeepVariant.output_vcf'], str)
        assert os.path.exists(outputs['GiraffeDeepVariant.output_vcf'])

    @slow
    @needs_singularity_or_docker
    def test_giraffe(self):
        """Test if Giraffe runs. This could take 12 minutes. Also we scale it down but it still demands lots of memory."""
        # TODO: enable test if nvidia-container-runtime and Singularity are installed but Docker isn't.

        json_dir = self._createTempDir()
        base_uri = 'https://raw.githubusercontent.com/vgteam/vg_wdl/65dd739aae765f5c4dedd14f2e42d5a263f9267a'
        wdl_file = f"{base_uri}/workflows/giraffe.wdl"
        json_file = f"{base_uri}/params/giraffe.json"

        result_json = subprocess.check_output(
            self.base_command + [wdl_file, json_file, '-o', self.output_dir, '--outputDialect', 'miniwdl', '--scale',
                                 '0.1'])
        result = json.loads(result_json)

        # Expect MiniWDL-style output with a designated "dir"
        assert 'dir' in result
        assert isinstance(result['dir'], str)
        out_dir = result['dir']

        assert 'outputs' in result
        assert isinstance(result['outputs'], dict)
        outputs = result['outputs']

        # Expect a BAM file to have been written
        assert 'Giraffe.output_bam' in outputs
        assert isinstance(outputs['Giraffe.output_bam'], str)
        assert os.path.exists(outputs['Giraffe.output_bam'])

    @needs_singularity_or_docker
    @needs_google_storage
    def test_gs_uri(self):
        """Test if Toil can access Google Storage URIs."""
        wdl = os.path.abspath('src/toil/test/wdl/md5sum/md5sum.1.0.wdl')
        json_file = os.path.abspath('src/toil/test/wdl/md5sum/md5sum-gs.json')

        result_json = subprocess.check_output(self.base_command + [wdl, json_file, '-o', self.output_dir, '--logDebug'])
        result = json.loads(result_json)

        assert 'ga4ghMd5.value' in result
        assert isinstance(result['ga4ghMd5.value'], str)
        assert os.path.exists(result['ga4ghMd5.value'])
        assert os.path.basename(result['ga4ghMd5.value']) == 'md5sum.txt'

    def test_coalesce(self):
        """
        Test if WDLSectionJob can coalesce WDL decls.

        White box test; will need to be changed or removed if the WDL interpreter changes.
        """

        # Set up data structures for our fake workflow graph to pull from.
        # This has all decl-type nodes
        all_decls: Set[str] = set()
        # And this has all transitive dependencies for all nodes.
        all_deps: Dict[str, Set[str]] = {}

        def mock_is_decl(self: Any, node_id: str) -> bool:
            """
            Replacement function to determine if a node is a decl or not.
            """
            return node_id in all_decls

        def mock_get_transitive_dependencies(self: Any, node_id: str) -> Set[str]:
            """
            Replacement function to get all the transitive dependencies of a node.
            """
            return all_deps[node_id]

        # These are the only two methods coalesce_nodes calls, so we can
        # replace them to ensure our graph is used without actually needing to
        # make any WDL objects for it.
        #
        # If that changes, the test will need to change! Maybe then it will be
        # worth extracting a base type for this interface.
        with patch.object(WDLWorkflowGraph, 'is_decl', mock_is_decl):
            with patch.object(WDLWorkflowGraph, 'get_transitive_dependencies', mock_get_transitive_dependencies):
                with self.subTest(msg="Two unrelated decls can coalesce"):
                    # Set up two unrelated decls
                    all_decls = {"decl1", "decl2"}
                    all_deps = {
                        "decl1": set(),
                        "decl2": set()
                    }

                    result = WDLSectionJob.coalesce_nodes(["decl1", "decl2"], WDLWorkflowGraph([]))

                    # Make sure they coalesced
                    assert len(result) == 1
                    assert "decl1" in result[0]
                    assert "decl2" in result[0]

                with self.subTest(msg="A decl will not coalesce with a non-decl"):
                    all_decls = {"decl"}
                    all_deps = {
                        "decl": set(),
                        "nondecl": set()
                    }

                    result = WDLSectionJob.coalesce_nodes(["decl", "nondecl"], WDLWorkflowGraph([]))

                    assert len(result) == 2
                    assert len(result[0]) == 1
                    assert len(result[1]) == 1

                with self.subTest(msg="Two adjacent decls with a common dependency can coalesce"):
                    all_decls = {"decl1", "decl2"}
                    all_deps = {
                        "decl1": {"base"},
                        "decl2": {"base"},
                        "base": set()
                    }

                    result = WDLSectionJob.coalesce_nodes(["base", "decl1", "decl2"], WDLWorkflowGraph([]))

                    assert len(result) == 2
                    assert "base" in result[0]
                    assert "decl1" in result[1]
                    assert "decl2" in result[1]

                with self.subTest(msg="Two adjacent decls with different dependencies will not coalesce"):
                    all_decls = {"decl1", "decl2"}
                    all_deps = {
                        "decl1": {"base"},
                        "decl2": set(),
                        "base": set()
                    }

                    result = WDLSectionJob.coalesce_nodes(["base", "decl1", "decl2"], WDLWorkflowGraph([]))

                    assert len(result) == 3
                    assert "base" in result[0]

                with self.subTest(msg="Two adjacent decls with different successors will coalesce"):
                    all_decls = {"decl1", "decl2"}
                    all_deps = {
                        "decl1": set(),
                        "decl2": set(),
                        "successor": {"decl2"}
                    }

                    result = WDLSectionJob.coalesce_nodes(["decl1", "decl2", "successor"], WDLWorkflowGraph([]))

                    assert len(result) == 2
                    assert "decl1" in result[0]
                    assert "decl2" in result[0]
                    assert "successor" in result[1]


@integrative
@slow
@pytest.mark.timeout(600)
class WDLKubernetesClusterTest(AbstractClusterTest):
    """
    Ensure WDL works on the Kubernetes batchsystem.
    """

    def __init__(self, name):
        super().__init__(name)
        self.clusterName = 'wdl-integration-test-' + str(uuid4())
        # t2.medium is the minimum t2 instance that permits Kubernetes
        self.leaderNodeType = "t2.medium"
        self.instanceTypes = ["t2.medium"]
        self.clusterType = "kubernetes"

    def setUp(self) -> None:
        super().setUp()
        self.jobStore = f'aws:{self.awsRegion()}:wdl-test-{uuid4()}'

    def launchCluster(self) -> None:
        self.createClusterUtil(args=['--leaderStorage', str(self.requestedLeaderStorage),
                                     '--nodeTypes', ",".join(self.instanceTypes),
                                     '-w', ",".join(self.numWorkers),
                                     '--nodeStorage', str(self.requestedLeaderStorage)])

    def test_wdl_kubernetes_cluster(self):
        """
        Test that a wdl workflow works on a kubernetes cluster. Launches a cluster with 1 worker. This runs a wdl
        workflow that performs an image pull on the worker.
        :return:
        """
        self.numWorkers = "1"
        self.requestedLeaderStorage = 30
        # create the cluster
        self.launchCluster()
        # get leader
        self.cluster = cluster_factory(
            provisioner="aws", zone=self.zone, clusterName=self.clusterName
        )
        self.leader = self.cluster.getLeader()

        url = "https://github.com/DataBiosphere/wdl-conformance-tests.git"
        commit = "09b9659cd01473e836738a2e0dd205df0adb49c5"
        wdl_dir = "wdl_conformance_tests"

        # get the wdl-conformance-tests repo to get WDL tasks to run
        self.sshUtil([
            "bash",
            "-c",
            f"git clone {url} {wdl_dir} && cd {wdl_dir} && git checkout {commit}"
        ])

        # run on kubernetes batchsystem
        toil_options = ['--batchSystem=kubernetes',
                        f"--jobstore={self.jobStore}"]

        # run WDL workflow that will run singularity
        test_options = [f"tests/md5sum/md5sum.wdl", f"tests/md5sum/md5sum.json"]
        self.sshUtil([
            "bash",
            "-c",
            f"cd {wdl_dir} && toil-wdl-runner {' '.join(test_options)} {' '.join(toil_options)}"])


if __name__ == "__main__":
    unittest.main()  # run all tests



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import subprocess
import tempfile
import time
from abc import abstractmethod
from inspect import getsource
from textwrap import dedent
from uuid import uuid4

import pytest

from toil.provisioners import cluster_factory
from toil.provisioners.aws.awsProvisioner import AWSProvisioner
from toil.test import (ToilTest,
                       integrative,
                       needs_aws_ec2,
                       needs_fetchable_appliance,
                       needs_mesos,
                       slow,
                       timeLimit)
from toil.test.provisioners.clusterTest import AbstractClusterTest
from toil.version import exactPython

log = logging.getLogger(__name__)


class AWSProvisionerBenchTest(ToilTest):
    """
    Tests for the AWS provisioner that don't actually provision anything.
    """

    # Needs to talk to EC2 for image discovery
    @needs_aws_ec2
    def test_AMI_finding(self):
        for zone in ['us-west-2a', 'eu-central-1a', 'sa-east-1b']:
            provisioner = AWSProvisioner('fakename', 'mesos', zone, 10000, None, None)
            ami = provisioner._discoverAMI()
            # Make sure we got an AMI and it looks plausible
            assert(ami.startswith('ami-'))

    @needs_aws_ec2
    def test_read_write_global_files(self):
        """
        Make sure the `_write_file_to_cloud()` and `_read_file_from_cloud()`
        functions of the AWS provisioner work as intended.
        """
        provisioner = AWSProvisioner(f'aws-provisioner-test-{uuid4()}', 'mesos', 'us-west-2a', 50, None, None)
        key = 'config/test.txt'
        contents = b"Hello, this is a test."

        try:
            url = provisioner._write_file_to_cloud(key, contents=contents)
            self.assertTrue(url.startswith("s3://"))

            self.assertEqual(contents, provisioner._read_file_from_cloud(key))
        finally:
            # the cluster was never launched, but we need to clean up the s3 bucket
            provisioner.destroyCluster()


@needs_aws_ec2
@needs_fetchable_appliance
@slow
@integrative
class AbstractAWSAutoscaleTest(AbstractClusterTest):
    def __init__(self, methodName):
        super().__init__(methodName=methodName)
        self.instanceTypes = ["m5a.large"]
        self.clusterName = 'aws-provisioner-test-' + str(uuid4())
        self.numWorkers = ['2']
        self.numSamples = 2
        self.spotBid = 0.15
        # We can't dump our user script right in /tmp or /home, because hot
        # deploy refuses to zip up those whole directories. So we make sure to
        # have a subdirectory to upload the script to.
        self.scriptDir = '/tmp/t'
        # Where should we put our virtualenv?
        self.venvDir = '/tmp/venv'
        # Where should we put our data to work on?
        # Must exist in the Toil container; the leader will try to rsync to it
        # (for the SSE key) and not create it.
        self.dataDir = '/tmp'
        # What filename should we use for our script (without path)?
        # Can be changed by derived tests.
        self.scriptName = 'test_script.py'

    def script(self):
        """
        Return the full path to the user script on the leader.
        """
        return os.path.join(self.scriptDir, self.scriptName)

    def data(self, filename):
        """
        Return the full path to the data file with the given name on the leader.
        """
        return os.path.join(self.dataDir, filename)

    def rsyncUtil(self, src, dest):
        subprocess.check_call(['toil', 'rsync-cluster', '--insecure', '-p=aws', '-z', self.zone, self.clusterName] + [src, dest])

    def getRootVolID(self):
        instances = self.cluster._get_nodes_in_cluster()
        instances.sort(key=lambda x: x.launch_time)
        leader = instances[0]  # assume leader was launched first

        from boto.ec2.blockdevicemapping import BlockDeviceType
        rootBlockDevice = leader.block_device_mapping["/dev/xvda"]
        assert isinstance(rootBlockDevice, BlockDeviceType)
        return rootBlockDevice.volume_id

    @abstractmethod
    def _getScript(self):
        """Download the test script needed by the inheriting unit test class."""
        raise NotImplementedError()

    def putScript(self, content: str):
        """
        Helper method for _getScript to inject a script file at the configured script path, from text.
        """
        cluster = cluster_factory(provisioner='aws', zone=self.zone, clusterName=self.clusterName)
        leader = cluster.getLeader()

        self.sshUtil(['mkdir', '-p', self.scriptDir])

        with tempfile.NamedTemporaryFile(mode='w') as t:
            # use appliance ssh method instead of sshutil so we can specify input param
            t.write(content)
            # This works to make writes visible on non-Windows
            t.flush()
            leader.injectFile(t.name, self.script(), 'toil_leader')


    @abstractmethod
    def _runScript(self, toilOptions):
        """
        Modify the provided Toil options to suit the test Toil script, then run the script with
        those arguments.

        :param toilOptions: List of Toil command line arguments. This list may need to be
               modified to suit the test script's requirements.
        """
        raise NotImplementedError()

    def _test(self, preemptibleJobs=False):
        """Does the work of the testing.  Many features' tests are thrown in here in no particular order."""
        # Make a cluster
        self.launchCluster()
        # get the leader so we know the IP address - we don't need to wait since create cluster
        # already insures the leader is running
        self.cluster = cluster_factory(provisioner='aws', zone=self.zone, clusterName=self.clusterName)
        self.leader = self.cluster.getLeader()
        self.sshUtil(['mkdir', '-p', self.scriptDir])
        self.sshUtil(['mkdir', '-p', self.dataDir])

        assert len(self.cluster._getRoleNames()) == 1
        # --never-download prevents silent upgrades to pip, wheel and setuptools
        venv_command = ['virtualenv', '--system-site-packages', '--python', exactPython, '--never-download', self.venvDir]
        self.sshUtil(venv_command)

        log.info('Set up script...')
        self._getScript()

        toilOptions = [self.jobStore,
                       '--workDir=/var/lib/toil',
                       '--clean=always',
                       '--retryCount=2',
                       '--logDebug',
                       '--logFile=' + os.path.join(self.scriptDir, 'sort.log')
                       ]

        if preemptibleJobs:
            toilOptions.extend(['--defaultPreemptible'])

        log.info('Run script...')
        self._runScript(toilOptions)

        assert len(self.cluster._getRoleNames()) == 1

        from boto.exception import EC2ResponseError
        volumeID = self.getRootVolID()
        self.cluster.destroyCluster()
        for attempt in range(6):
            # https://github.com/BD2KGenomics/toil/issues/1567
            # retry this for up to 1 minute until the volume disappears
            try:
                self.boto2_ec2.get_all_volumes(volume_ids=[volumeID])
                time.sleep(10)
            except EC2ResponseError as e:
                if e.status == 400 and 'InvalidVolume.NotFound' in e.code:
                    break
                else:
                    raise
        else:
            self.fail('Volume with ID %s was not cleaned up properly' % volumeID)

        assert len(self.cluster._getRoleNames()) == 0


@integrative
@needs_mesos
@pytest.mark.timeout(1800)
class AWSAutoscaleTest(AbstractAWSAutoscaleTest):
    def __init__(self, name):
        super().__init__(name)
        self.clusterName = 'provisioner-test-' + str(uuid4())
        self.requestedLeaderStorage = 80
        self.scriptName = 'sort.py'

    def setUp(self):
        super().setUp()
        self.jobStore = f'aws:{self.awsRegion()}:autoscale-{uuid4()}'

    def _getScript(self):
        fileToSort = os.path.join(os.getcwd(), str(uuid4()))
        with open(fileToSort, 'w') as f:
            # Fixme: making this file larger causes the test to hang
            f.write('01234567890123456789012345678901')
        self.rsyncUtil(os.path.join(self._projectRootPath(), 'src/toil/test/sort/sort.py'), ':' + self.script())
        self.rsyncUtil(fileToSort, ':' + self.data('sortFile'))
        os.unlink(fileToSort)

    def _runScript(self, toilOptions):
        toilOptions.extend(['--provisioner=aws', '--batchSystem=mesos',
                            '--nodeTypes=' + ",".join(self.instanceTypes),
                            '--maxNodes=' + ",".join(self.numWorkers)])
        runCommand = [self.python(), self.script(), '--fileToSort=' + self.data('sortFile'), '--sseKey=' + self.data('sortFile')]
        runCommand.extend(toilOptions)
        self.sshUtil(runCommand)

    def launchCluster(self):
        # add arguments to test that we can specify leader storage
        self.createClusterUtil(args=['--leaderStorage', str(self.requestedLeaderStorage)])

    def getRootVolID(self):
        """
        Adds in test to check that EBS volume is build with adequate size.
        Otherwise is functionally equivalent to parent.
        :return: volumeID
        """
        volumeID = super().getRootVolID()
        rootVolume = self.boto2_ec2.get_all_volumes(volume_ids=[volumeID])[0]
        # test that the leader is given adequate storage
        self.assertGreaterEqual(rootVolume.size, self.requestedLeaderStorage)
        return volumeID

    @integrative
    @needs_aws_ec2
    def testAutoScale(self):
        self.instanceTypes = ["m5a.large"]
        self.numWorkers = ['2']
        self._test()

    @integrative
    @needs_aws_ec2
    def testSpotAutoScale(self):
        self.instanceTypes = ["m5a.large:%f" % self.spotBid]
        self.numWorkers = ['2']
        self._test(preemptibleJobs=True)

    @integrative
    @needs_aws_ec2
    def testSpotAutoScaleBalancingTypes(self):
        self.instanceTypes = ["m5.large/m5a.large:%f" % self.spotBid]
        self.numWorkers = ['2']
        self._test(preemptibleJobs=True)


@integrative
@needs_mesos
@pytest.mark.timeout(2400)
class AWSStaticAutoscaleTest(AWSAutoscaleTest):
    """Runs the tests on a statically provisioned cluster with autoscaling enabled."""
    def __init__(self, name):
        super().__init__(name)
        self.requestedNodeStorage = 20

    def launchCluster(self):
        from boto.ec2.blockdevicemapping import BlockDeviceType

        from toil.lib.ec2 import wait_instances_running
        self.createClusterUtil(args=['--leaderStorage', str(self.requestedLeaderStorage),
                                     '--nodeTypes', ",".join(self.instanceTypes),
                                     '-w', ",".join(self.numWorkers),
                                     '--nodeStorage', str(self.requestedLeaderStorage)])

        self.cluster = cluster_factory(provisioner='aws', zone=self.zone, clusterName=self.clusterName)
        # We need to wait a little bit here because the workers might not be
        # visible to EC2 read requests immediately after the create returns,
        # which is the last thing that starting the cluster does.
        time.sleep(10)
        nodes = self.cluster._get_nodes_in_cluster()
        nodes.sort(key=lambda x: x.launch_time)
        # assuming that leader is first
        workers = nodes[1:]
        # test that two worker nodes were created
        self.assertEqual(2, len(workers))
        # test that workers have expected storage size
        # just use the first worker
        worker = workers[0]
        worker = next(wait_instances_running(self.boto2_ec2, [worker]))
        rootBlockDevice = worker.block_device_mapping["/dev/xvda"]
        self.assertTrue(isinstance(rootBlockDevice, BlockDeviceType))
        rootVolume = self.boto2_ec2.get_all_volumes(volume_ids=[rootBlockDevice.volume_id])[0]
        self.assertGreaterEqual(rootVolume.size, self.requestedNodeStorage)

    def _runScript(self, toilOptions):
        # Autoscale even though we have static nodes
        toilOptions.extend(['--provisioner=aws', '--batchSystem=mesos',
                            '--nodeTypes=' + ",".join(self.instanceTypes),
                            '--maxNodes=' + ",".join(self.numWorkers)])
        runCommand = [self.python(), self.script(), '--fileToSort=' + self.data('sortFile')]
        runCommand.extend(toilOptions)
        self.sshUtil(runCommand)


@integrative
@pytest.mark.timeout(1200)
class AWSManagedAutoscaleTest(AWSAutoscaleTest):
    """Runs the tests on a self-scaling Kubernetes cluster."""
    def __init__(self, name):
        super().__init__(name)
        self.requestedNodeStorage = 20

    def launchCluster(self):
        from boto.ec2.blockdevicemapping import BlockDeviceType  # noqa

        from toil.lib.ec2 import wait_instances_running  # noqa
        self.createClusterUtil(args=['--leaderStorage', str(self.requestedLeaderStorage),
                                     '--nodeTypes', ",".join(self.instanceTypes),
                                     '--workers', ",".join([f'0-{c}' for c in self.numWorkers]),
                                     '--nodeStorage', str(self.requestedLeaderStorage),
                                     '--clusterType', 'kubernetes'])

        self.cluster = cluster_factory(provisioner='aws', zone=self.zone, clusterName=self.clusterName)

    def _runScript(self, toilOptions):
        # Don't use the provisioner, and use Kubernetes instead of Mesos
        toilOptions.extend(['--batchSystem=kubernetes'])
        runCommand = [self.python(), self.script(), '--fileToSort=' + self.data('sortFile')]
        runCommand.extend(toilOptions)
        self.sshUtil(runCommand)


@integrative
@needs_mesos
@pytest.mark.timeout(1200)
class AWSAutoscaleTestMultipleNodeTypes(AbstractAWSAutoscaleTest):
    def __init__(self, name):
        super().__init__(name)
        self.clusterName = 'provisioner-test-' + str(uuid4())

    def setUp(self):
        super().setUp()
        self.jobStore = f'aws:{self.awsRegion()}:autoscale-{uuid4()}'

    def _getScript(self):
        sseKeyFile = os.path.join(os.getcwd(), 'keyFile')
        with open(sseKeyFile, 'w') as f:
            f.write('01234567890123456789012345678901')
        self.rsyncUtil(os.path.join(self._projectRootPath(), 'src/toil/test/sort/sort.py'), ':' + self.script())
        self.rsyncUtil(sseKeyFile, ':' + self.data('keyFile'))
        os.unlink(sseKeyFile)

    def _runScript(self, toilOptions):
        # Set memory requirements so that sort jobs can be run
        # on small instances, but merge jobs must be run on large
        # instances
        toilOptions.extend(['--provisioner=aws', '--batchSystem=mesos',
                            '--nodeTypes=' + ",".join(self.instanceTypes),
                            '--maxNodes=' + ",".join(self.numWorkers)])
        runCommand = [self.python(), self.script(), '--fileToSort=/home/s3am/bin/asadmin', '--sortMemory=0.6G', '--mergeMemory=3.0G']
        runCommand.extend(toilOptions)
        runCommand.append('--sseKey=' + self.data('keyFile'))
        self.sshUtil(runCommand)

    @integrative
    @needs_aws_ec2
    def testAutoScale(self):
        self.instanceTypes = ["t2.small", "m5a.large"]
        self.numWorkers = ['2', '1']
        self._test()


@integrative
@needs_mesos
@pytest.mark.timeout(1200)
class AWSRestartTest(AbstractAWSAutoscaleTest):
    """This test insures autoscaling works on a restarted Toil run."""
    def __init__(self, name):
        super().__init__(name)
        self.clusterName = 'restart-test-' + str(uuid4())
        self.scriptName = 'restartScript.py'

    def setUp(self):
        super().setUp()
        self.instanceTypes = ['t2.small']
        self.numWorkers = ['1']
        self.jobStore = f'aws:{self.awsRegion()}:restart-{uuid4()}'

    def _getScript(self):
        def restartScript():
            import os

            from configargparse import ArgumentParser

            from toil.job import Job

            def f0(job):
                if 'FAIL' in os.environ:
                    raise RuntimeError('failed on purpose')

            if __name__ == '__main__':
                parser = ArgumentParser()
                Job.Runner.addToilOptions(parser)
                options = parser.parse_args()
                rootJob = Job.wrapJobFn(f0, cores=0.5, memory='50 M', disk='50 M')
                Job.Runner.startToil(rootJob, options)

        script = dedent('\n'.join(getsource(restartScript).split('\n')[1:]))
        self.putScript(script)

    def _runScript(self, toilOptions):
        # Use the provisioner in the workflow
        toilOptions.extend(['--provisioner=aws', '--batchSystem=mesos',
                            '--nodeTypes=' + ",".join(self.instanceTypes),
                            '--maxNodes=' + ",".join(self.numWorkers)])
        # clean = onSuccess
        disallowedOptions = ['--clean=always', '--retryCount=2']
        newOptions = [option for option in toilOptions if option not in disallowedOptions]
        try:
            # include a default memory - on restart the minimum memory requirement is the default, usually 2 GB
            command = [self.python(), self.script(), '--setEnv', 'FAIL=true', '--defaultMemory=50000000']
            command.extend(newOptions)
            self.sshUtil(command)
        except subprocess.CalledProcessError:
            pass
        else:
            self.fail('Command succeeded when we expected failure')
        with timeLimit(600):
            command = [self.python(), self.script(), '--restart', '--defaultMemory=50000000']
            command.extend(toilOptions)
            self.sshUtil(command)

    def testAutoScaledCluster(self):
        self._test()


@integrative
@needs_mesos
@pytest.mark.timeout(1200)
class PreemptibleDeficitCompensationTest(AbstractAWSAutoscaleTest):
    def __init__(self, name):
        super().__init__(name)
        self.clusterName = 'deficit-test-' + str(uuid4())
        self.scriptName = 'userScript.py'

    def setUp(self):
        super().setUp()
        self.instanceTypes = ['m5a.large:0.01', "m5a.large"]  # instance needs to be available on the spot market
        self.numWorkers = ['1', '1']
        self.jobStore = f'aws:{self.awsRegion()}:deficit-{uuid4()}'

    def test(self):
        self._test(preemptibleJobs=True)

    def _getScript(self):
        def userScript():
            from toil.common import Toil
            from toil.job import Job

            # Because this is the only job in the pipeline and because it is preemptible,
            # there will be no non-preemptible jobs. The non-preemptible scaler will therefore
            # not request any nodes initially. And since we made it impossible for the
            # preemptible scaler to allocate any nodes (using an abnormally low spot bid),
            # we will observe a deficit of preemptible nodes that the non-preemptible scaler will
            # compensate for by spinning up non-preemptible nodes instead.
            #
            def job(job, disk='10M', cores=1, memory='10M', preemptible=True):
                pass

            if __name__ == '__main__':
                options = Job.Runner.getDefaultArgumentParser().parse_args()
                with Toil(options) as toil:
                    if toil.config.restart:
                        toil.restart()
                    else:
                        toil.start(Job.wrapJobFn(job))

        script = dedent('\n'.join(getsource(userScript).split('\n')[1:]))
        self.putScript(script)

    def _runScript(self, toilOptions):
        toilOptions.extend(['--provisioner=aws', '--batchSystem=mesos',
                            '--nodeTypes=' + ",".join(self.instanceTypes),
                            '--maxNodes=' + ",".join(self.numWorkers)])
        toilOptions.extend(['--preemptibleCompensation=1.0'])
        command = [self.python(), self.script()]
        command.extend(toilOptions)
        self.sshUtil(command)
